
"""
File name:		batchgcd.py

Description:	Implementation of the batch-GCD algorithm for finding factors
				of a list of numbers

Author:			Eli Howey
				Written for Final Project
				MATH 56, Spring 2014

"""

import time



def producttree(X):
	"""
	producttree(X)
	Returns the product tree of a list of numbers X, where:
		- leaves are the numbers in X,
		- root is the product of all numbers in X,
		- each intermediate node is the product of its (at most) two children

	Note: The tree is represented layer by layer, from bottom to top.
	Code credited to Dan Bernstein
	"""
	tree = [X]

	while len(X) > 1:
		# Multiplies each pair of elements in the list each iteration
		X = [ prod(X[ 2*i:2*(i+1) ]) for i in range((len(X)+1)/2) ]
		tree.append(X)

	return tree



def remainders(n,X):
	"""
	remainders(n,X)
	Calculates the remainders of n modulo each element of the list X

	Code credited to Dan Bernstein
	"""
	result = [n]
	T = producttree(X)

	for t in reversed(T):
		# For each layer of the tree, calculate the remainder of n modulo
		# each of the children composing each element in the layer, using:
		#		n mod X[0] = (n (mod X[0]*X[1])) (mod X[0])
		#		n mod X[1] = (n (mod X[0]*X[1])) (mod X[1])
		result = [ result[floor(i/2)] % t[i] for i in range(len(t)) ]

	return result



def batchgcd_simple(X):
	"""
	batchgcd_simple(X):
	determines which elements share primes with which others by calculating
	pairwise GCDs

	Element X[i] shares primes with some other element X[j] iff the i-th
	element of batchgcd_simple(X) is greater than 1
	"""
	P = prod(X)

	X2 = []
	for x in X:
		X2.append(x*x)

	# Calculate remainders of product modulo X_i^2
	R = remainders(P, X2)

	# gcd(X_i, prod(X)/X_i) = gcd(X_i, (prod(X) mod X_i^2)/X_i)
	G = []
	for i in range(len(X)):
		g = gcd(X[i], R[i]/X[i])
		G.append(g)

	return G



def main():
	print "\n" + "=" * 80
	print "Testing batchgcd"
	print "=" * 80 + "\n"

	p1, p2, p3 = 29, 53, 73			# To demonstrate what GCD looks like

	X = [
		range(1,6),								# For basic understanding
		list(primes(50)),						# All primes, no shared factors
		[47*p2, 37*91, p1*p3, 61*23, 43*p1],	# Two keys will factor
		[p2*p3, 37*91, p1*p3, 61*23, p1*p2]		# Unfactorizable cycle
		]


	for i in range(len(X)):
		keys = X[i]

		print "Test", i+1
		print "-" * 80
		print "Keys:", keys

		G = batchgcd_simple(keys)
		print "batchgcd output:", G, "\n"

		prime = [False] * len(G)	# Keep track of prime GCDs

		# Record keys that share factors
		s = ""
		for i in range(len(G)):
			if G[i] != 1:
				s += str(keys[i]) + ", "

				# If GCD is prime, we can factor the key!
				# Note: This may be difficult for real-world examples
				if is_prime(G[i]): prime[i] = True

		if s == "":
			print "No keys share any factors with any other keys\n\n"
		else:
			print "The following keys share at least one factor with some other key(s)"
			print s[:-2] + "\n"

			# Print factorizations of all factorizable keys
			no_prime_gcds = True

			for i in range(len(G)):
				if prime[i]:
					no_prime_gcds = False
					print keys[i], "=", G[i], "x", keys[i] / G[i]

			# If there are no prime GCDs, we can't factor anything
			if no_prime_gcds:
				print "Unable to factor any keys"
			print "\n"

	print "=" * 80
	print "End of test"
	print "=" * 80 + "\n"


main()