"""
File name:		quadratic_sieve.py

Description:	Implementation of the quadratic sieve factorization method

Author:			Eli Howey
				Written for Final Project
				MATH 56, Spring 2014

"""

import time

DEBUG = False		# Set to True to see printout info
VERBOSE = False		# Set to True for sieve printout info



def quadratic_root(p, a):
	"""
	quadratic_root(p, a)
	Finds an x such that x^2 = a (mod p)

	Assumes p odd prime, and Legendre symbol (a/p) = 1
	"""
	# Check simplest cases: p = 3,5,7 (mod 8)
	a %= p
	if p % 8 == 3 or p % 8 == 7:
		return pow(a, (p+1)/4, p)

	if p % 8 == 5:
		x = pow(a, (p+3)/8, p)
		c = pow(x,2,p)
		if c != a:
			x = (x * pow(2, (p-1/4))) % p
		return x

	# Final case: p = 1 (mod 8)
	d = randint(2,p-1)
	if kronecker(d,p) != -1:
		d = randint(2,p-1)

	t = p-1
	s = 0
	while t % 2 == 0:
		t /= 2
		s += 1

	A = pow(a,t,p)
	# D = pow(d,t,p)
	D = d % p
	m = 0

	for i in range(s):
		prod = pow(A * (D ** m), 2 ** (s-1-i), p)
		if (prod == -1):
			m += 2 ** i

	return ((pow(a, (t+1)/2) * pow(D, m/2))) % p



def quadratic_sieve(n):
	"""
	quadratic_sieve(n)
	Quadratic sieve factorization algorithm [Pomerance 266]
	"""
	# =========================================================================
	# Part 1. Setup
	# =========================================================================

	# Optimal smoothness bound
	L = exp(sqrt(ln(n) * ln(ln(n))))
	B = ceil(L ** (.5))

	if DEBUG: print "Smoothness bound:", B

	fb = [2]		# Always sieve with 2, even if (n/2) != 1
	roots = [1]		# Quadratic residue of 2 mod n

	# Factor base only needs primes that are quadratic residues mod n
	fb += [x for x in list(primes(B)) if kronecker(n,x) == 1 and x > 2]
	pi_fb = len(fb)

	if DEBUG: print "Factor base:", fb

	# Find quadratic residues of each prime in factor base
	for i in range(1,pi_fb):
		# For p_i > 2, quadratic roots mod p_i are at indices 2i-1, 2i
		r = quadratic_root(fb[i], n)
		roots.append(r)
		roots.append(-r)

	if DEBUG: print "Roots:", roots, "\n"

	# =========================================================================
	# Part 2. Sieving
	# =========================================================================

	SIEVE_SIZE = 32					# Arbitrary
	sieve_count = 0					# Number of sieve subintervals performed
	smooth = []						# Array of B-smooth numbers
	x0 = int(ceil(sqrt(n)))			# Initial value of x

	x = x0
	C = (x * x) - n

	# Create sub-sieves repeatedly until we have enough smooth numbers
	while len(smooth) <= pi_fb:
		# Populate a sub-sieve with candidates
		sieve = []
		for i in range(SIEVE_SIZE):
			sieve.append(C)
			C  = (C + (2 * x + 1)) % n
			x += 1
		if DEBUG and VERBOSE:
			print "Reducing sieve", sieve_count, ":\n", sieve, "\n"

		# Copy the sub-sieve to manipulate it
		s_cpy = list(sieve)

		# Sieve for p = 2 separately (since it only has 1 residue)
		shift = x0 % 2
		a = (roots[0] - shift + sieve_count * SIEVE_SIZE) % 2	# Start index

		while a < SIEVE_SIZE:
			while s_cpy[a] % 2 == 0:
				s_cpy[a] /= 2
			a += 2

		# Sieve by each odd prime in the factor base
		for i in range(1, pi_fb):
			p = fb[i]

			# Calculate initial sieve indices whose value p divides
			shift = x0 % p
			a1 = (int(roots[2*i-1]) - shift + sieve_count * SIEVE_SIZE) % p
			a2 = (int(roots[2*i]) - shift + sieve_count * SIEVE_SIZE) % p

			# Sieving by first residue:
			while a1 < SIEVE_SIZE:
				while s_cpy[a1] % p == 0:
					s_cpy[a1] /= p
				a1 += p

			# Sieving by second residue:
			while a2 < SIEVE_SIZE:
				while s_cpy[a2] % p == 0:
					s_cpy[a2] /= p
				a2 += p

		# Record all numbers x that have smooth x^2 % n
		for i in range(SIEVE_SIZE):
			if s_cpy[i] == 1 and sieve[i] != 1:
				pair = (x0 + i + sieve_count * SIEVE_SIZE, sieve[i])
				smooth.append(pair)
				if DEBUG and VERBOSE:
					print "Found smooth number:", pair[1], "\n"

		sieve_count += 1

	if DEBUG: print "Smooth values:", smooth, "\n"

	# =========================================================================
	# Part 3. Linear Algebra
	# =========================================================================

	# Create exponent vectors from each smooth number
	exps = []
	for t in smooth:
		s = t[1]			# The smooth number
		v = [0] * pi_fb		# Exponent vector

		for i in range(pi_fb):
			p = fb[i]
			while s % p == 0:
				s /= p
				v[i] += 1

		exps.append(v)

	if DEBUG: print "Exponent vectors:", exps

	# Find linear combinations of smooth numbers to produce a square
	A = matrix(GF(2),len(exps),len(fb), exps)
	null = A.left_kernel().basis()

	if DEBUG and VERBOSE: print "Null basis:", null, "\n"

	# For each null vector, construct u and v such that u^2 = v^2 mod n
	for nvec in null:
		if DEBUG and VERBOSE: print "Trying null vector", nvec

		u = 1
		for i in range(len(smooth)):
			if nvec[i] == 1:
				# u = product of x_i's
				u = (u * smooth[i][0]) % n

		if DEBUG and VERBOSE: print "u =", u

		# Create an exponent vector for v
		v_exps = [0] * pi_fb
		for i in range(len(smooth)):
			if nvec[i] == 1:
				for j in range(pi_fb):
					v_exps[j] += exps[i][j]
					v_exps[j] /= 2

		if DEBUG and VERBOSE: print "Exponent vector of v:", v_exps

		# Calculate v from its exponent vector
		v = 1
		for i in range(pi_fb):
			p_pow = pow(fb[i], v_exps[i], n)
			v = (v * p_pow) % n

		if DEBUG and VERBOSE: print "v =", v, "\n"

		g1 = gcd(u+v, n)
		g2 = gcd(u-v, n)

		if g1 != 1 and g2 != 1:
			if DEBUG and not VERBOSE:
				print "Valid null vector:", nvec
				print "u =", u
				print "v =", v, "\n"

			return (g1, g2)

	# If no null vector gives nontrivial factors, return 1 and n
	# (This shouldn't ever happen, though!)
	if DEBUG and not VERBOSE:
		print "No valid null vectors found"
	return (1,n)



def main():
	N =[
		8051,		# 83 * 97
		90283,		# 137 * 659
		832147,		# 857 * 971
		7863409,	# 1489 * 5281
		40903063	# 5953 * 6871
		]

	T = []

	for n in N:
		print "\nAttempting to factor", n

		t = time.time()
		Q = quadratic_sieve(n)
		T.append(time.time() - t)

		print Q, "\n"
		print "-" * 75

	print T


main()