"""
File name:		sieve.py

Description:	Implementation of the Sieve of Eratosthenes

Author:			Eli Howey
				Written for Final Project
				MATH 56, Spring 2014

"""

from math import ceil, sqrt, log
import matplotlib.pyplot as plt
import time



def sieve(n):
	"""
	sieve(n)
	Returns a list of all primes less than n, collected using the Sieve of
		Eratosthenes
	"""
	# Sieve runs from 2 to n, but more convenient from 0 to n
	sieve = [True] * n
	sieve[0] = sieve[1] = False

	# Run through elements in the sieve, crossing off multiples
	max = int(ceil(sqrt(n))) + 1
	for i in range(max):
		if sieve[i]:
			j = 2 * i
			while j < n:
				sieve[j] = False
				j += i

	# Collect the primes from the sieve
	primes = []
	for i in range(n):
		if sieve[i]:
			primes.append(i)

	return primes



def factor_sieve(n):
	"""
	factor_sieve(n)
	Returns a list called sieve, where sieve[i] is:
		1 if i is prime
		the smallest prime factor of i if i is composite
		0 if i = 0,1 (neither prime nor composite)

	Can be used to factor any integer less than n through repeated division
		by the relevant entries in sieve
	"""
	n = int(n)
	sieve = [1] * n				# Assume all numbers prime
	sieve[0] = sieve[1] = 0

	max = int(ceil(sqrt(n))) + 1
	for i in range(max):
		# If number is prime, record that prime as the smallest factor of its multiples (if applicable)
		if sieve[i] == 1:
			j = 2 * i
			while j < n:
				# If a multiple does not have a smaller factor,
				# record the factor
				if sieve[j] == 1:
					sieve[j] = i
				j += i

	return sieve



def trial_divide(n, sieve):
	"""
	trial_divide(n, sieve)
	Returns a list containing the prime factorization of n, constructed
		by trial division

	Note: Having the factor sieve as an input allows for the sieve to be
		calculated only once, instead of every trial division. The sieve must
		range from 1 to n.
	"""
	n = int(n)
	if len(sieve) <= n:
		return "Error: trial_divide: sieve does not contain n"

	factors = []				# Factorization of n

	# Record the smallest prime factor of n, and divide n by that factor,
	# until n is prime
	while sieve[n] != 1:
		f = sieve[n]
		factors.append(f)
		n /= f
	factors.append(n)		# Append that prime factor

	return (sorted(factors), len(factors))



def main():
	N = [pow(10,i) for i in range(1,7)]		# Upper bounds
	T = []									# Runtimes for each bound

	for n in N:
		t = time.time()
		sieve(n)
		T.append(time.time() - t)

	plt.loglog(N,T)
	plt.xlabel('n')
	plt.ylabel('Runtime (s)')
	plt.show()


main()

